import { sleep } from '../utils/eventLoop';
import { StringifiableValue } from '../utils/serializer';
import { CacheEntry, createCachedDataProvider } from './cache';
import { createKeyValueStorage } from './inMemoryDB';

type ValueType = Record<string, StringifiableValue>;

describe('createCachedDataProvider', () => {
    it('always fetches remote data when ttl is zero, for sequential calls', async () => {
        const calledWith: unknown[] = [];
        const getValue = createCachedDataProvider({
            cacheStorage: createKeyValueStorage<CacheEntry<ValueType>>(0),
            getNewValue: (key) => {
                calledWith.push(key);
                return Promise.resolve(key);
            },
            ttlMs: 0,
        });

        expect(await getValue({ key: 1 })).toEqual({ key: 1 });
        await sleep(10); // to make sure that Date.now() does not always return the same value
        expect(await getValue({ key: 2 })).toEqual({ key: 2 });
        await sleep(10);
        expect(await getValue({ key: 1 })).toEqual({ key: 1 });
        await sleep(10);
        expect(await getValue({ key: 1 })).toEqual({ key: 1 });
        await sleep(10);
        expect(await getValue({ key: 3 })).toEqual({ key: 3 });

        expect(calledWith).toEqual([
            { key: 1 },
            { key: 2 },
            { key: 1 },
            { key: 1 },
            { key: 3 },
        ]);
    });

    it('never refetches remote data when ttl is 1 hour, for sequential calls', async () => {
        const calledWith: unknown[] = [];
        const getValue = createCachedDataProvider({
            cacheStorage: createKeyValueStorage<CacheEntry<ValueType>>(0),
            getNewValue: (key) => {
                calledWith.push(key);
                return Promise.resolve(key);
            },
            ttlMs: 3_600_000,
        });

        expect(await getValue({ key: 1 })).toEqual({ key: 1 });
        await sleep(10);
        expect(await getValue({ key: 2 })).toEqual({ key: 2 });
        await sleep(10);
        expect(await getValue({ key: 1 })).toEqual({ key: 1 });
        await sleep(10);
        expect(await getValue({ key: 1 })).toEqual({ key: 1 });
        await sleep(10);
        expect(await getValue({ key: 3 })).toEqual({ key: 3 });

        expect(calledWith).toEqual([{ key: 1 }, { key: 2 }, { key: 3 }]);
    });

    it('refetches remote data when ttl is 1 hour but cache is reset', async () => {
        const cacheStorage = createKeyValueStorage<CacheEntry<ValueType>>(0);
        const calledWith: unknown[] = [];
        const getValue = createCachedDataProvider({
            cacheStorage,
            getNewValue: (key) => {
                calledWith.push(key);
                return Promise.resolve(key);
            },
            ttlMs: 3_600_000,
        });

        expect(await getValue({ key: 1 })).toEqual({ key: 1 });
        await cacheStorage.clear();
        expect(await getValue({ key: 1 })).toEqual({ key: 1 });

        expect(calledWith).toEqual([{ key: 1 }, { key: 1 }]);
    });

    it('refetches remote data after 200ms when ttl is 100ms', async () => {
        const calledWith: unknown[] = [];
        const getValue = createCachedDataProvider({
            cacheStorage: createKeyValueStorage<CacheEntry<ValueType>>(0),
            getNewValue: (key: ValueType) => {
                const callNumber = calledWith.length;
                calledWith.push(key);
                return Promise.resolve({ ...key, callNumber });
            },
            ttlMs: 100,
        });

        expect(await getValue({ key: 1 })).toEqual({ key: 1, callNumber: 0 });
        expect(await getValue({ key: 2 })).toEqual({ key: 2, callNumber: 1 });
        expect(await getValue({ key: 1 })).toEqual({ key: 1, callNumber: 0 });
        expect(await getValue({ key: 2 })).toEqual({ key: 2, callNumber: 1 });
        await sleep(200);
        expect(await getValue({ key: 1 })).toEqual({ key: 1, callNumber: 2 });
        expect(await getValue({ key: 2 })).toEqual({ key: 2, callNumber: 3 });
        expect(await getValue({ key: 1 })).toEqual({ key: 1, callNumber: 2 });
        expect(await getValue({ key: 2 })).toEqual({ key: 2, callNumber: 3 });

        expect(calledWith).toEqual([
            { key: 1 },
            { key: 2 },
            { key: 1 },
            { key: 2 },
        ]);
    });

    it('never refetches remote data even when ttl is zero, for concurrent calls', async () => {
        const calledWith: unknown[] = [];
        const getValue = createCachedDataProvider({
            // We need noticeable cache latency for deterministic behavior, so that it will always throttle
            // instead of switching between throttling and caching at random
            cacheStorage: createKeyValueStorage<CacheEntry<ValueType>>(10),
            getNewValue: (key: ValueType) => {
                const callNumber = calledWith.length;
                calledWith.push(key);
                return Promise.resolve({ ...key, callNumber });
            },
            ttlMs: 0,
        });

        expect(
            await Promise.all([
                getValue({ key: 1 }),
                getValue({ key: 2 }),
                getValue({ key: 1 }),
                getValue({ key: 1 }),
                getValue({ key: 3 }),
            ]),
        ).toEqual([
            { key: 1, callNumber: 0 },
            { key: 2, callNumber: 1 },
            { key: 1, callNumber: 0 },
            { key: 1, callNumber: 0 },
            { key: 3, callNumber: 2 },
        ]);

        expect(calledWith).toEqual([{ key: 1 }, { key: 2 }, { key: 3 }]);
    });
});
