import { throttle } from '../utils/throttle';
import { type StringifiableValue, createSerializer } from '../utils/serializer';
import type { ClearableKeyValueStorage } from './types';

type DataProvider<TKey, TValue> = (key: TKey) => Promise<TValue>;

export type CacheEntry<TValue> = {
    expirationDate: number;
    cachedValue: TValue;
};

export const createCachedDataProvider = <
    TKey extends StringifiableValue,
    TValue,
>({
    cacheStorage,
    getNewValue,
    ttlMs,
}: {
    cacheStorage: ClearableKeyValueStorage<string, CacheEntry<TValue>>;
    getNewValue: DataProvider<TKey, TValue>;
    ttlMs: number;
}): DataProvider<TKey, TValue> => {
    const keySerializer = createSerializer<TKey>();
    const unsafeGet = async (key: TKey) => {
        const cacheEntry = await cacheStorage.get(keySerializer.stringify(key));
        if (cacheEntry.found && cacheEntry.value.expirationDate >= Date.now()) {
            return cacheEntry.value.cachedValue;
        }

        const newValue = await getNewValue(key);
        await cacheStorage.set(keySerializer.stringify(key), {
            expirationDate: Date.now() + ttlMs,
            cachedValue: newValue,
        });

        return newValue;
    };

    return throttle(unsafeGet);
};
