import { nextTick, sleep } from './eventLoop';
import { throttle } from './throttle';

const createDeferred = async () => {
    let promise: Promise<void> | undefined = undefined;
    const { settle } = await new Promise<{ settle: () => void }>((resolve) => {
        promise = new Promise((settle) => {
            resolve({ settle });
        });
    });

    return {
        // `promise` is guaranteed to be defined here, because we only reach this code after `resolve({ settle })` was called,
        // which cannot happen in earlier event loop iteration than the one where `promise` is assigned.
        // ESLint correctly complains here because this is an anti-pattern that makes it hard to automatically reason about the code.
        // We only do async crimes here to write deterministic tests for low-level async `throttle` function.
        // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
        promise: promise!,
        settle,
    };
};

describe('accepts arguments and returns results for sequential calls with partially repeated arguments', () => {
    it('calls sync function and returns value', async () => {
        const calledWith: unknown[] = [];
        const f = throttle(
            (...args: unknown[]) =>
                new Promise((resolve) => {
                    calledWith.push(args);
                    resolve(args);
                }),
        );

        expect(await f(1, 2, 3)).toEqual([1, 2, 3]);
        expect(await f(1, 2, 4)).toEqual([1, 2, 4]);
        expect(await f(1, 2, 4)).toEqual([1, 2, 4]);
        expect(await f(5)).toEqual([5]);
        expect(await f(1, 2, 3)).toEqual([1, 2, 3]);

        expect(calledWith).toEqual([
            [1, 2, 3],
            [1, 2, 4],
            [1, 2, 4],
            [5],
            [1, 2, 3],
        ]);
    });

    it('calls async function with nextTick and returns results', async () => {
        const calledWith: unknown[] = [];
        const f = throttle(async (...args: unknown[]) => {
            calledWith.push(args);
            await nextTick();
            return args;
        });

        expect(await f(1, 2, 3)).toEqual([1, 2, 3]);
        expect(await f(1, 2, 4)).toEqual([1, 2, 4]);
        expect(await f(1, 2, 4)).toEqual([1, 2, 4]);
        expect(await f(5)).toEqual([5]);
        expect(await f(1, 2, 3)).toEqual([1, 2, 3]);

        expect(calledWith).toEqual([
            [1, 2, 3],
            [1, 2, 4],
            [1, 2, 4],
            [5],
            [1, 2, 3],
        ]);
    });

    it('calls async function with small sleep and returns results', async () => {
        const calledWith: unknown[] = [];
        const f = throttle(async (...args: unknown[]) => {
            calledWith.push(args);
            await sleep(50);
            return args;
        });

        expect(await f(1, 2, 3)).toEqual([1, 2, 3]);
        expect(await f(1, 2, 4)).toEqual([1, 2, 4]);
        expect(await f(1, 2, 4)).toEqual([1, 2, 4]);
        expect(await f(5)).toEqual([5]);
        expect(await f(1, 2, 3)).toEqual([1, 2, 3]);

        expect(calledWith).toEqual([
            [1, 2, 3],
            [1, 2, 4],
            [1, 2, 4],
            [5],
            [1, 2, 3],
        ]);
    });
});

describe('accepts arguments and returns results for concurrent calls with different arguments', () => {
    it('calls sync function and returns value', async () => {
        const calledWith: unknown[] = [];
        const f = throttle(
            (...args: unknown[]) =>
                new Promise((resolve) => {
                    calledWith.push(args);
                    resolve(args);
                }),
        );

        expect(await Promise.all([f(1, 2, 3), f(1, 2, 4), f(5)])).toEqual([
            [1, 2, 3],
            [1, 2, 4],
            [5],
        ]);

        expect(calledWith).toEqual([[1, 2, 3], [1, 2, 4], [5]]);
    });

    it('calls async function with nextTick and returns results', async () => {
        const calledWith: unknown[] = [];
        const f = throttle(async (...args: unknown[]) => {
            calledWith.push(args);
            await nextTick();
            return args;
        });

        expect(await Promise.all([f(1, 2, 3), f(1, 2, 4), f(5)])).toEqual([
            [1, 2, 3],
            [1, 2, 4],
            [5],
        ]);

        expect(calledWith).toEqual([[1, 2, 3], [1, 2, 4], [5]]);
    });

    it('calls async function with small sleep and returns results', async () => {
        const calledWith: unknown[] = [];
        const f = throttle(async (...args: unknown[]) => {
            calledWith.push(args);
            await sleep(50);
            return args;
        });

        expect(await Promise.all([f(1, 2, 3), f(1, 2, 4), f(5)])).toEqual([
            [1, 2, 3],
            [1, 2, 4],
            [5],
        ]);

        expect(calledWith).toEqual([[1, 2, 3], [1, 2, 4], [5]]);
    });
});

describe('accepts arguments and returns results for concurrent calls with partially repeated arguments', () => {
    it('calls sync function and returns value', async () => {
        const calledWith: unknown[] = [];
        const f = throttle(
            (...args: unknown[]) =>
                new Promise((resolve) => {
                    calledWith.push(args);
                    resolve(args);
                }),
        );

        expect(
            await Promise.all([
                f(1, 2, 3),
                f(1, 2, 4),
                f(1, 2, 4),
                f(5),
                f(1, 2, 3),
            ]),
        ).toEqual([[1, 2, 3], [1, 2, 4], [1, 2, 4], [5], [1, 2, 3]]);

        expect(calledWith).toEqual([[1, 2, 3], [1, 2, 4], [5]]);
    });

    it('calls async function with nextTick and returns results', async () => {
        const calledWith: unknown[] = [];
        const f = throttle(async (...args: unknown[]) => {
            calledWith.push(args);
            await nextTick();
            return args;
        });

        expect(
            await Promise.all([
                f(1, 2, 3),
                f(1, 2, 4),
                f(1, 2, 4),
                f(5),
                f(1, 2, 3),
            ]),
        ).toEqual([[1, 2, 3], [1, 2, 4], [1, 2, 4], [5], [1, 2, 3]]);

        expect(calledWith).toEqual([[1, 2, 3], [1, 2, 4], [5]]);
    });

    it('calls async function with small sleep and returns results', async () => {
        const calledWith: unknown[] = [];
        const f = throttle(async (...args: unknown[]) => {
            calledWith.push(args);
            await sleep(50);
            return args;
        });

        expect(
            await Promise.all([
                f(1, 2, 3),
                f(1, 2, 4),
                f(1, 2, 4),
                f(5),
                f(1, 2, 3),
            ]),
        ).toEqual([[1, 2, 3], [1, 2, 4], [1, 2, 4], [5], [1, 2, 3]]);

        expect(calledWith).toEqual([[1, 2, 3], [1, 2, 4], [5]]);
    });
});

describe('waits for the wrapped function to resolve prior to calling it again', () => {
    it('for the same arguments', async () => {
        let deferred = await createDeferred();
        const calledWith: unknown[] = [];
        const f = throttle(async (...args: unknown[]) => {
            calledWith.push(args);
            const callNumber = calledWith.length;
            await deferred.promise;
            return { args, callNumber };
        });

        const promises = [];
        promises.push(f(1, 2, 3));
        promises.push(f(1, 2, 3));
        promises.push(f(1, 2, 3));
        deferred.settle();
        await Promise.all(promises);
        deferred = await createDeferred();
        promises.push(f(1, 2, 3));
        promises.push(f(1, 2, 3));
        deferred.settle();

        expect(await Promise.all(promises)).toEqual([
            { args: [1, 2, 3], callNumber: 1 },
            { args: [1, 2, 3], callNumber: 1 },
            { args: [1, 2, 3], callNumber: 1 },
            { args: [1, 2, 3], callNumber: 2 },
            { args: [1, 2, 3], callNumber: 2 },
        ]);

        expect(calledWith).toEqual([
            [1, 2, 3],
            [1, 2, 3],
        ]);
    });

    it('for different arguments', async () => {
        let deferred = await createDeferred();
        const calledWith: unknown[] = [];
        const f = throttle(async (...args: unknown[]) => {
            calledWith.push(args);
            const callNumber = calledWith.length;
            await deferred.promise;
            return { args, callNumber };
        });

        const promises = [];
        promises.push(f(1, 2, 3));
        promises.push(f(1, 2, 4));
        promises.push(f(1, 2, 3));
        deferred.settle();
        await Promise.all(promises);
        deferred = await createDeferred();
        promises.push(f(1, 2, 4));
        promises.push(f(1, 2, 3));
        deferred.settle();

        expect(await Promise.all(promises)).toEqual([
            { args: [1, 2, 3], callNumber: 1 },
            { args: [1, 2, 4], callNumber: 2 },
            { args: [1, 2, 3], callNumber: 1 },
            { args: [1, 2, 4], callNumber: 3 },
            { args: [1, 2, 3], callNumber: 4 },
        ]);

        expect(calledWith).toEqual([
            [1, 2, 3],
            [1, 2, 4],
            [1, 2, 4],
            [1, 2, 3],
        ]);
    });
});
